﻿/*
 * 
 *          DremDOS
 *          Main Terminal Classs (Terminal.cs)
 * 
 *          Contains the main backend logic for Consoles. It includes the command parser and some basic commands.
 *
 */

using System;
using System.Collections.Generic;
using System.Text;
using Cosmos.System.Graphics;
using System.Drawing;

namespace DremDOS.Core.Terminal
{
    class Terminal
    {
        private VGATerminal LinkedVGATerminal;
        private Kernel kernel;
        private char[] CommandBuffer = new char[1000];
        public bool ShowOutput = false;
        private int CommandBufferCurrentLocation = 0;

        private Pen CurrentPen = new Pen(Color.White);
        private string CurrentDirectory = @"0:\";

        private bool ConsoleMode;
        private bool RunCommandReady;
        public Terminal(bool IsInConsoleMode, Kernel k)
        {
            kernel = k;
            Initialize(IsInConsoleMode);
        }

        private void Initialize(bool IsInConsoleMode)
        {
            RunCommandReady = true;
            ConsoleMode = IsInConsoleMode;
            if (ConsoleMode)
            {
                ShowOutput = true;
            }
            InitializeBuffer(); // JUST in case Initialize has some extra stuff that needs done later. J U S T in case.
        }
        private void InitializeBuffer()
        {
            CommandBufferCurrentLocation = 0;
            for (int i = 0; i < CommandBuffer.Length; i++)
            {
                CommandBuffer[i] = '\0';
            }
        }
        public void SetPen(Pen Color) { CurrentPen = Color; }
        public void SetLinkedVGATerminal(VGATerminal Term)
        {
            LinkedVGATerminal = Term;
            ShowOutput = true;
        }
        public void WriteToBuffer(char c) {
            CommandBuffer[CommandBufferCurrentLocation] = c;
            CommandBufferCurrentLocation+=1;
            if (!ConsoleMode) { if (CommandBufferCurrentLocation == 0) { LinkedVGATerminal.CanBackspace = false; } else { LinkedVGATerminal.CanBackspace = true; } }

            if(ConsoleMode)
            {
                AppendText("" + c);
            }
        }
        public void ClearBuffer() { InitializeBuffer(); LinkedVGATerminal.CanBackspace = false; }
        public void BufferBackspace() {
            if (CommandBufferCurrentLocation != 0) {
                CommandBuffer[CommandBufferCurrentLocation - 1] = '\0';
                CommandBufferCurrentLocation--;
                if (!ConsoleMode) { if (CommandBufferCurrentLocation == 0) { LinkedVGATerminal.CanBackspace = false; } else { LinkedVGATerminal.CanBackspace = true; } }
            } else {
                if (!ConsoleMode) { LinkedVGATerminal.CanBackspace = false; }
            }
        }

        public void EnterPressed()
        {
            if(RunCommandReady)
            {
                RunCommand();
            }
        }
        public void RunCommand() {
            string temp = "";
            char nullchar = '\0';
            for(int i = 0; i < CommandBuffer.Length; i++)
            {
                if(CommandBuffer[i] == nullchar) { break; }
                temp += CommandBuffer[i].ToString();
            }
            //LinkedVGATerminal.AppendText("You entered: " + temp + "\n");
            RunCommand(temp);

            ClearBuffer();
        }

        private void AppendText(string Text)
        {
            if(ConsoleMode)
            {
                Console.Write(Text);
            } else
            {
                LinkedVGATerminal.AppendText(Text);
            }
            return;
        }
        private void AppendText(string Text, Pen Color)
        {
            if (ConsoleMode)
            {
                Console.Write(Text);
            }
            else
            {
                LinkedVGATerminal.AppendText(Text, Color);
            }
            return;
        } // I don't recommend using this as it makes single console mode ugly.

        private void AppendText(string Text, string Color)
        {
            if (ConsoleMode)
            {
                ConsoleColor PreviousForeground = Console.ForegroundColor;
                ConsoleColor Foreground = StringToConsole(Color);
                Console.ForegroundColor = Foreground;
                Console.Write(Text);
                Console.ForegroundColor = PreviousForeground;

            }
            else
            {
                Pen ColorPen = StringToPen(Color);
                LinkedVGATerminal.AppendText(Text, ColorPen);
            }
            return;
        } // The above "AppendText" methods adds an ambiguity layer that lets the OS choose what to output the text to. Important for single console mode.

        private ConsoleColor StringToConsole(string ColorIn)
        {
            ConsoleColor Color;
            Color = ConsoleColor.White;

            // TODO: Unfugg this now that switch statements exist.
            if (ColorIn == "Black") {
                Color = ConsoleColor.Black;
            } else if (ColorIn == "DarkBlue") {
                Color = ConsoleColor.DarkBlue;
            } else if (ColorIn == "DarkGreen") {
                Color = ConsoleColor.DarkGreen;
            } else if (ColorIn == "DarkCyan") {
                Color = ConsoleColor.DarkCyan;
            } else if (ColorIn == "DarkRed") {
                Color = ConsoleColor.DarkRed;
            } else if (ColorIn == "DarkMagenta") {
                Color = ConsoleColor.DarkMagenta;
            } else if (ColorIn == "DarkYellow") {
                Color = ConsoleColor.DarkYellow;
            } else if (ColorIn == "Gray") {
                Color = ConsoleColor.Gray;
            } else if (ColorIn == "DarkGray") {
                Color = ConsoleColor.DarkGray;
            } else if (ColorIn == "Blue") {
                Color = ConsoleColor.Blue;
            } else if (ColorIn == "Green") {
                Color = ConsoleColor.Green;
            } else if (ColorIn == "Cyan") {
                Color = ConsoleColor.Cyan;
            } else if (ColorIn == "Red") {
                Color = ConsoleColor.Red;
            } else if (ColorIn == "Magenta") {
                Color = ConsoleColor.Magenta;
            } else if (ColorIn == "Yellow") {
                Color = ConsoleColor.Yellow;
            } else { // Consider having it throw an exception if the color is invalid.
                Color = ConsoleColor.White;
            } // This hurts to look at

            return Color;
        }

        private Pen StringToPen(string ColorIn)
        {
            Pen PenColor;
            //PenColor = new Pen(Color.FromName(ColorIn));

            // This is not the full list of available colors, but it matches the available colors to Console.
            // This is a temporary stop-gap until System.Drawing.Color.FromName finally works.
            switch(ColorIn)
            {
                case "Black":
                    PenColor = new Pen(Color.Black);
                    break;
                case "DarkBlue":
                    PenColor = new Pen(Color.DarkBlue);
                    break;
                case "DarkGreen":
                    PenColor = new Pen(Color.DarkGreen);
                    break;
                case "DarkCyan":
                    PenColor = new Pen(Color.DarkCyan);
                    break;
                case "DarkRed":
                    PenColor = new Pen(Color.DarkRed);
                    break;
                case "DarkMagenta":
                    PenColor = new Pen(Color.DarkMagenta);
                    break;
                case "DarkYellow":
                    //PenColor = new Pen(ColorTranslator.FromHtml("#cc7722"));
                    PenColor = new Pen(Color.Yellow); // DarkYellow == Yellow for now.
                    break;
                case "Gray":
                    PenColor = new Pen(Color.Gray);
                    break;
                case "DarkGray":
                    PenColor = new Pen(Color.DarkGray);
                    break;
                case "Blue":
                    PenColor = new Pen(Color.Blue);
                    break;
                case "Green":
                    PenColor = new Pen(Color.Green);
                    break;
                case "Cyan":
                    PenColor = new Pen(Color.Cyan);
                    break;
                case "Red":
                    PenColor = new Pen(Color.Red);
                    break;
                case "Magenta":
                    PenColor = new Pen(Color.Magenta);
                    break;
                case "Yellow":
                    PenColor = new Pen(Color.Yellow);
                    break;
                default: // Consider having it throw an exception if the color is invalid.
                    // Check if user wanted a hex value
                    // This is a pretty poor and inefficient way of doing this but... eh I don't care.
                    // The hex conversion should really be a different function.
                    /*if (ColorIn[0] == '0') // This can be uncommented when ColorTranslator is implemented.
                    {
                        // More validation is required.
                        PenColor = new Pen(ColorTranslator.FromHtml(ColorIn));
                        break;
                    }*/
                    PenColor = new Pen(Color.White);
                    break;
            }

            return PenColor;
        }

        public void RunCommand(string cmd)
        {
            string[] arguments = GetArguments(cmd);
            //RunCommand(arguments);
            Command();
            //Console.Beep(1000, 1000);
        }

        public void RunCommand(string[] arguments)
        {
            if (arguments[0] == "command")
            {
                //Console.Beep(1500, 1000);
                Command();
            }
            else if (arguments[0] == "checktime")
            {
                if(ShowOutput) { AppendText(CheckTime()); }
            }
            else if (arguments[0] == "help")
            {
                if (ShowOutput) { Help(arguments); }
            }
            else if (arguments[0] == "beep")
            {
                Console.Beep();
            }
            else if (arguments[0] == "clear")
            {
                if(ConsoleMode)
                {
                    Console.Clear();
                } else
                {
                    //LinkedVGATerminal.Clear(); Implement later
                }
            }
            else
            {
                if (ShowOutput) { AppendText(@"ERROR: Bad Command!"); }
            }

            if (ShowOutput)
            {
                AppendText("\n" + CurrentDirectory + " >");
            }

            ClearBuffer();
        }

        private void Command() {
            if (ShowOutput)
            {
                AppendText(@"                                           _..._" + '\n', "Blue"); // Blue
                AppendText(@"    ___                ___  ____  ____   ", "Red"); // Red
                AppendText(@".:     '." + '\n', "DarkBlue"); // Dark Blue
                AppendText(@"   / _ \_______ __ _  / _ \/ __ \/ __/  ", "Yellow"); // Yellow
                AppendText(@":::::     :" + '\n', new Pen(Color.DarkBlue)); // Dark Blue
                AppendText(@"  / // / __/ -_)  ' \/ // / /_/ /\ \    ", "Green");  // Green
                AppendText(@":::::::   :" + '\n', "DarkMagenta"); // Dark Magenta
                AppendText(@" /____/_/  \__/_/_/_/____/\____/___/    ", "Blue");   // Blue
                AppendText(@"`::::::::.'" + '\n', "DarkMagenta");    // Dark Magenta
                AppendText(@"                                          `':::''" + '\n', "Magenta"); // Magenta
                AppendText(kernel._OS_NAME + " " + kernel._OS_VERSION_FULL + "\n");
            }
        }

        private string CheckTime()
        {
            // Get the time/date and output it
            string time = DateTime.Now.ToString();
            //LinkedVGATerminal.AppendText("{0}", time);
            return time;
        }

        private void Help(string[] arguments)
        {
            // Tests if the arguments has the correct length
            if (arguments.Length >= 2)
            {
                // If command == [valid command], output information on the command
                if (arguments[1] == "help")
                {
                    AppendText("help - help menu or get help on a command\n");
                    AppendText("help <command> - get help on a specific command");
                }
                else if (arguments[1] == "command")
                {
                    AppendText("command - outputs information about DremDOS");
                }
                else if (arguments[1] == "beep")
                {
                    AppendText("beep - make the PC speaker go beep!\n");
                    AppendText("beep <frequency> <time> - make a more specific beep");
                }
                else if (arguments[1] == "cls")
                {
                    AppendText("cls - clear the screen");
                }
                else if (arguments[1] == "dir")
                {
                    AppendText("dir - display all files and folders in the current directory");
                }
                else if (arguments[1] == "cd")
                {
                    AppendText("cd - change the directory\n");
                    AppendText("cd <directory> - change to <directory>\n");
                    AppendText("  You can either specify a specific directory or relative directory.\n");
                    AppendText("cd .. - go up one directory");
                }
                else if (arguments[1] == "drivelist")
                {
                    AppendText("drivelist - list all currently attached drives, their filesystems, and other information");
                }
                else if (arguments[1] == "checktime")
                {
                    AppendText("checktime - check the time");
                }
                else if (arguments[1] == "calculate")
                {
                    AppendText("calculate - a basic calculator\n");
                    AppendText("calculate <int> <operator> <int>");
                }
                else if (arguments[1] == "shutdown")
                {
                    AppendText("shutdown - shut down the computer\n");
                    AppendText("shutdown /r - restart the computer");
                }
                else if (arguments[1] == "kitty")
                {
                    AppendText("kitty - read and write files\n");
                    AppendText("kitty <file> [/r] [/w] [/nf] <text>\n");
                    AppendText("Example: kitty foo.txt /r\n");
                    AppendText("  Read foo.txt and output it to the screen\n");
                    AppendText("Example: kitty foo.txt /w \"bar\"\n");
                    AppendText("  Write \"bar\" to foo.txt (without quotes)");
                    AppendText("Example: kitty foo.txt /nf\n");
                    AppendText("  Create a new file named \"foo.txt\" in the current directory");
                }
                else if (arguments[1] == "rm")
                {
                    AppendText("rm - remove a file from the drive\n");
                    AppendText("rm <file>");
                }
                else if (arguments[1] == "rmdir")
                {
                    AppendText("rmdir - remove an empty directory\n");
                    AppendText("rmdir <directory>");
                }
                else if (arguments[1] == "mkdir")
                {
                    AppendText("mkdir - create an empty directory\n");
                    AppendText("mkdir <directory>");
                }
                else if (arguments[1] == "copy")
                {
                    AppendText("copy - copy a file\n");
                    AppendText("copy <source> <destination>\n");
                    AppendText("Note: Copy is currently very limited.");
                }
                else if (arguments[1] == "move")
                {
                    AppendText("move - move a file\n");
                    AppendText("move <source> <destination>\n");
                    AppendText("Note: currently doesn't work");
                }
                else if (arguments[1] == "initgui")
                {
                    AppendText("\nNotice: Not yet implemented!\n");
                    //AppendText("Usage: initgui [resolution] <color depth>");
                    AppendText("Usage: initgui [resolution]\n");
                    AppendText("\n");
                    AppendText("Available Resolutions:\n");
                    //AppendText("       640x480     800x600    1024x768");
                    AppendText("       800x600     1024x768    1280x720\n");
                    AppendText("     1920x1080\n");
                    /*AppendText("");
                    AppendText("Available Color Depths:");
                    AppendText("             4           8           16");
                    AppendText("            24          32");*/
                    //AppendText("                   800x600    1024x768");
                    AppendText("\n");
                    AppendText("Example: initgui 800x600\n");
                    AppendText("\n");
                    //AppendText("Usage: initgui");
                    //AppendText("Initializes the experimental GUI.\nNote: No longer supports any resolution except 1024x768.");
                }
                else if (arguments[1] == "textop")
                {
                    AppendText("\nNotice: Not yet implemented!\n");
                    AppendText("Usage: textop [console size]\n");
                    AppendText("\n");
                    AppendText("Available console sizes:\n");
                    //AppendText("       640x480     800x600    1024x768");
                    //AppendText("      1280x720    1366x768    1920x1080");
                    AppendText("       80x25\n");
                    AppendText("\n");
                    AppendText("Example: textop 80x25\n");
                    AppendText("\n");
                }
                else if (arguments[1] == "ver")
                {
                    AppendText("ver - display version");
                }
                else // Else, display the default help menu if the command requested in the first argument isn't valid
                {
                    AppendText("DremDOS Help Menu\n\n");
                    AppendText("command - ouputs information about DremDOS\n");
                    AppendText("beep - make the PC speaker go beep!\n");
                    AppendText("cls - clear the screen\n");
                    AppendText("dir - display all files and folders in the current directory\n");
                    AppendText("cd - change the directory\n");
                    AppendText("drivelist - list all currently attached drives and their filesystems\n");
                    AppendText("checktime - check the time\n");
                    AppendText("calculate - a basic calculator\n");
                    AppendText("shutdown - shut down the computer\n");
                    AppendText("kitty - read and write files\n");
                    AppendText("rm - remove a file from the drive\n");
                    AppendText("rmdir - remove an empty directory\n");
                    AppendText("mkdir - create an empty directory\n");
                    AppendText("copy - copy a file\n");
                    AppendText("move - move a file\n");
                    AppendText("initgui - start the experimental desktop environment\n");
                    AppendText("textop - start the experimental Textop environment\n");
                    AppendText("ver - display version\n");
                    AppendText("help - help menu or get help on a command\n\n");
                    AppendText("Tip: Try help <command>");
                }
            }
            else // Default help menu
            {
                // Ack I'm tired I'll figure this out later
                AppendText("DremDOS Help Menu\n\n");
                AppendText("command - ouputs information about DremDOS\n");
                AppendText("beep - make the PC speaker go beep!\n");
                AppendText("cls - clear the screen\n");
                AppendText("dir - display all files and folders in the current directory\n");
                AppendText("cd - change the directory\n");
                AppendText("drivelist - list all currently attached drives and their filesystems\n");
                AppendText("checktime - check the time\n");
                AppendText("calculate - a basic calculator\n");
                AppendText("shutdown - shut down the computer\n");
                AppendText("kitty - read and write files\n");
                AppendText("rm - remove a file from the drive\n");
                AppendText("rmdir - remove an empty directory\n");
                AppendText("mkdir - create an empty directory\n");
                AppendText("copy - copy a file\n");
                AppendText("move - move a file\n");
                AppendText("initgui - start the experimental desktop environment\n");
                AppendText("textop - start the experimental Textop environment\n");
                AppendText("ver - display version\n");
                AppendText("help - help menu or get help on a command\n\n");
                AppendText("Tip: Try help <command>");
            }
        }

        /*private string TrimString(string str)
        {
            string temp = "";
            char[] charArray = temp.ToCharArray();

            for (int i = 0; i < charArray.Length; i++)
            {
                if(charArray[i] != ' ')
                {
                    temp += charArray[i];
                }
            }

            return temp;
        }*/

        // Props to Çağrı
        // https://stackoverflow.com/a/61112392/11500310
        private string[] GetArguments(string cmd)
        {
            List<string> result = new List<string>();
            var split1 = cmd.Split('"');
            for (int i = 0; i < split1.Length; i++)
            {
                split1[i] = split1[i].Trim();
                //split1[i] = TrimString(split1[i]); // string.Trim is no longer working.
            }

            for (int i = 0; i < split1.Length; i++)
            {
                if (i % 2 == 0)
                {
                    var split2 = split1[i].Split(' ');
                    foreach (var el in split2)
                    {
                        result.Add(el);
                    }
                }
                else
                {
                    result.Add(split1[i]);
                }

            }
            string[] arr = new string[result.Count];
            for (int i = 0; i < result.Count; i++)
            {
                arr[i] = result[i];
            }
            return arr;
        }
    }
}
